/* See LICENSE file for license details. */

#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdlib.h>
#ifdef USE_SHADOW
#include <shadow.h>
#endif
#include <string.h>
#include <unistd.h>

#include "../slock.h"

struct AuthState {
	char *hash;
};

void
auth_redraw(const struct State *const state)
{
}

void
auth_reset(const struct State *const state, char *const msg, const int fail)
{
}

void
auth_backspace(const struct State *const state)
{
}

void
auth_input(const struct State *const state, const char *const buf,
		const int len)
{
}

void
auth_init(struct State *const state)
{
	char *inputhash;
	const struct passwd *pw;
	int rv;

	/* Read password hash and then drop elevated privileges. */
	state->auth = malloc(sizeof(*state->auth));
	pw = getpwuid(getuid());
	check_val(pw != NULL, "getpwuid", strerror(errno));
	state->auth->hash = pw->pw_passwd;

#ifdef USE_SHADOW
	const struct spwd *sp;
	if (!strcmp(state->auth->hash, "x")) {
		sp = getspnam(pw->pw_name);
		check_val(sp != NULL, "getspnam", "cannot retrieve shadow "
			"entry. Make sure to suid or sgid.");
		state->auth->hash = sp->sp_pwdp;
	}
#else
#ifdef __OpenBSD__
	if (!strcmp(state->auth->hash, '*')) {
		pw = getpwuid_shadow(getuid());
		check_val(pw != NULL, "getpwuid_shadow", strerror(errno));
		state->auth->hash = pw->pw_passwd;
	}
#endif
#endif

	inputhash = crypt("", state->auth->hash);
	check_val(inputhash != NULL, "crypt", strerror(errno));
	rv = setgroups(0, NULL);
	check_val(rv == 0, "setgroups", strerror(errno));
	rv = setgid(getgid());
	check_val(rv == 0, "setgid", strerror(errno));
	rv = setuid(getuid());
	check_val(rv == 0, "setuid", strerror(errno));
}

int
auth_do(struct State *const state)
{
	char *inputhash;
	int rv;

	inputhash = crypt(state->input, state->auth->hash);
	if (!inputhash) {
		reset(state, strerror(errno), 1);
		return -1;
	}
	rv = strcmp(inputhash, state->auth->hash);
	explicit_bzero(inputhash, sizeof(inputhash));
	reset(state, "authentication failure", rv);
	return rv;
}

void
auth_close(const struct State *const state)
{
	explicit_bzero(state->auth->hash, sizeof(state->auth->hash));
}
